#!/bin/bash

# This script uses sglr to parse the program given as argument,
# and asfe to translate it to funcons.
#
# It should be run from Languages/Pico.
#
##########
# Command-line options:
#
# -n : Skip dumping of equations and parse table creation:
# This 'compilation' only needs to be done when a module 
# in Languages/Pico is modified. 
#
# -m : Specify ASF translation module
#
# -f : Source of translation
#
# -o : Target of translation

##########
# The output of this script may be redirected to a file:
# ../../process-pico.sh -n -f test.pico > test.fct


# Delete any temporary files on exit.
# trap 'rm -f TEMPFILE.tbl; rm -f stderrout.txt' EXIT

# Obtain Meta-Environment base directory by using location of asfsdf-meta
# according to the PATH.
METADIR=$(dirname `type -P asfsdf-meta`)/..

# By default, 'compile' the Pico specification. (Actually, dump equations,
# then use the asfe tool to rewrite a parsed term according to the dumped
# equations.)
# 
# Ideally, we would use the asfc tool to create a binary. However, on my
# computer, the generated binary omits layout in the generated SDF
# files. Using asfe generates files that are identical to those generated by
# using the ASF+SDF Meta-Environment.
COMPILE=true

# Process command-line options
while getopts m:npf:d:l:go: OPTION
do
  case "$OPTION" in
    m) MODULE="$OPTARG";;
    n) COMPILE=false;;
    f) FILE="$OPTARG";;
    o) OUTFILE="$OPTARG";;
  esac
done

if $COMPILE; then
  echo "***** Dumping equations:" >&2
  echo "Creating file $MODULE.eqs (This may take a few minutes)" >&2
  eqs-dump -m $MODULE -o $MODULE.eqs >&2
  
  echo >&2
  echo "***** Creating $MODULE parse table:" >&2
  echo "Creating file $MODULE.tbl" >&2
  sdf2table -c -p .:$METADIR/share/asf-library/library:$METADIR/share/sdf-library/library -m $MODULE -o $MODULE.tbl

  echo >&2
fi

  echo ">>>>> $FILE" >&2
  echo -n "Translating to funcons: " >&2
  PARSECSFTOSDF=$(sglr -p $MODULE.tbl -i "$FILE" -A -t 2> /dev/null)
  if [[ $? == "1" ]]; then
    echo "Parse Error!" >&2
    #echo $RESULT >&2
    #exit 1
  fi
  AMBRESULT=$(ambtracker <<< "${PARSECSFTOSDF}")
  if [[ "$AMBRESULT" != "" ]]; then
    echo "Ambiguity!"
    echo "$AMBRESULT"
    #exit 2
  fi
  RESULT=$(asfe -e $MODULE.eqs -t <<< "${PARSECSFTOSDF}" | unparsePT)
  
  echo "Done" >&2
  echo >&2

  echo "$RESULT" | tr -d '\n' | tr -d '\t' | #remove newlines and tabs
        sed 's_  *_ _g' | sed 's_, _,_g'  | #change space sequence to single space
        sed 's_^ __g' | sed 's_ $_\n_g' > "$OUTFILE" #remove leading space, change trailing space to newline

  
